%%
%% A theory of integers, loosely based on the Mural Integer theory.
%% =================================================
%%
%% The only primitives are 0, succ, pred, =< and is_int.
%%
%% From these primitives, we define '+', '*', '**', '/', mod, '<'.
%% Negative literals are represented using a prefix negation operator ('-').
%% To avoid redundancy, the following operators are not supported:
%%
%%	A - B		-- use A + (- B) instead
%%	A > B		-- use B < A instead (or not A =< B)
%%	A >= B		-- use B =< A instead (or not A < B)
%%

add_theory(arith).
inherit_theory(r).

abbreviation A > B === B < A.
abbreviation A >= B === B =< A.
abbreviation A - B === A + (- B).
abbreviation A:ints === int(A).

const(0).
const(1).

function int(_).
function succ(_).
function pred(_).

absolute_function 700 nonassoc '=<,le'(_,_).


axiom int_localplaces
=== A = B
    => int(A) = int(B).

%%
%% Adding the following axiom avoids the need for axioms about the
%% localness of succ and pred (which, if done carelessly, might
%% contradict the definition of succ in set theory).
%%
%% However, this axiom can be proved from localness properties of succ and
%% pred, plus the other axioms below.
%%
axiom succ_1_to_1
=== int(A) and int(B) 
    => (A = B <=> succ(A) = succ(B)).

axiom succ_localplaces
===  (A = B) => (succ(A) = succ(B)).

axiom pred_localplaces
===  (A = B) => (pred(A) = pred(B)).

%% Axioms (based on the Mural theory)
axiom int_bool
=== bool(int(A)).


axiom zero_int
=== int(0).

axiom succ_int
=== int(I) => int(succ(I)).

axiom pred_int
=== int(I) => int(pred(I)).

axiom pred_succ
=== int(I) => pred(succ(I)) = I.

axiom succ_pred
=== int(I) => succ(pred(I)) = I.

axiom le_bool
=== int(A) and int(B) => bool(A =< B).

axiom le_reflex
=== int(I) => I =< I.

axiom le_trans
=== int(I) and int(J) and int(K) 
    and I =< J and J =< K => I =< K.

axiom le_antisym
=== int(I) and int(J) => (I =< J and J =< I <=> I = J).

axiom int_total_order
=== int(I) and int(J) => I =< J or J =< I.

axiom le_mono_succ
=== int(I) and int(J) => (I =< J <=> succ(I) =< succ(J)).

axiom le_succ
===  int(I) => I =< succ(I).

axiom neq_succ
===  int(I) => I \= succ(I).

axiom int_induction
=== (is_local x P)
    and [0/x]*P
    and (all x  int(x) and 0 =< x and P => [succ(x)/x]*P)
    and (all x  int(x) and x =< 0 and P => [pred(x)/x]*P)
    => (all x int(x) => P).

%% This axiom seems to be necessary.
%% It serves a similar purpose to the third peano axiom.
%% (ie. all natural numbers are generated by 0 and succ).
%%
axiom succ_cases
=== int(A) and int(B) and A =< B and B =< succ(A)
    => A = B or B = succ(A).



%% Derived concepts
%% ================

%% not the ideal handling of non-integer arguments
function 700 nonassoc '<,lt'(A,B)
=== A =< B and A \= B.

%%
%% (due to the lack of recursive definitions, we declare the following
%% functions as primitive, then give each of their definitions as
%% a single axiom.
%%

constant nats.

axiom def_nats
=== X:nats = (X:ints and 0 =< X).

%% theorem nat_to_int
%% === A:nats => A:ints.
rerun(arith:nat_to_int).

absolute_function 500 leftassoc   '+,add'(_,_).
absolute_function 500 prefix      '-,neg'(_).
absolute_function 400 leftassoc   '*,mul'(_,_).
absolute_function 300 rightassoc  '**,exp'(_,_).

axiom def_add
=== int(A) and int(B)
    => A + B = if(B = 0, 
			A,
		if(B < 0, 
			pred(A + succ(B)),
		%% B > 0
			succ(A + pred(B)))).

%% It would be better to replace the following two axioms by a 
%% single recursive definition that does two pred's for each succ,
%% or vice versa.
%%

axiom neg_int
=== int(I) => int(- I).

axiom add_inv_r
=== int(I) => I + (- I) = 0.


axiom def_mul
=== int(A) and int(B)
    => A * B = if(B = 0, 
			0,
		if(B < 0, 
			(- A) + (A * succ(B)),
		%% B > 0
			A + (A * pred(B)))).

axiom def_exp
=== int(A) and int(B) and 0 =< B
    => A ** B = if(B = 0,  1,  A * (A ** pred(B))).


%% Some opening rules
openrule(succ(_),  [1], [], id, [= --> =], [succ(G) --> G]).
openrule(pred(_),  [1], [], id, [= --> =], [pred(G) --> G]).
openrule(int(_),   [1], [], id, [= --> =], [int(G)  --> G]).
openrule(A =< B,   [1], [], id, [= --> =], [G =< B  --> G]).
openrule(A =< B,   [2], [], id, [= --> =], [A =< G  --> G]).
openrule(A < B,    [1], [], id, [= --> =], [G < B   --> G]).
openrule(A < B,    [2], [], id, [= --> =], [A < G   --> G]).
openrule(A + B,    [1], [], id, [= --> =], [G + B   --> G]).
openrule(A + B,    [2], [], id, [= --> =], [A + G   --> G]).
openrule( -(A),    [1], [], id, [= --> =], [-(G)    --> G]).
openrule(A * B,    [1], [], id, [= --> =], [G * B   --> G]).
openrule(A * B,    [2], [], id, [= --> =], [A * G   --> G]).
openrule(A ** B,   [1], [], id, [= --> =], [G ** B  --> G]).
openrule(A ** B,   [2], [], id, [= --> =], [A ** G  --> G]).



rerun(arith : add_id_r).
rerun(arith : def_lt).
rerun(arith : lt_succ).
rerun(arith : eq_imp_le).
rerun(arith : def_le).
rerun(arith : le_not_lt).
rerun(arith : lt_not_le).
rerun(arith : lt_imp_le).
rerun(arith : eq_not_lt).
rerun(arith : eq_not_lt_rev).
rerun(arith : int_cases).
rerun(arith : lt_le_trans).
rerun(arith : le_lt_trans).
rerun(arith : lt_mono_succ).
rerun(arith : lt_pred).
rerun(arith : localsucc).
rerun(arith : localmul).
rerun(arith : localle).
rerun(arith : localexp).
rerun(arith : localpred).
rerun(arith : localneg).
rerun(arith : localint).
rerun(arith : localadd).

rerun(arith : succ_case).  
rerun(arith : pred_case).
rerun(arith : add_id_l).

rerun(arith : lt_mono_pred).
rerun(arith : le_mono_pred).
rerun(arith : pred_1_to_1).
rerun(arith : lt_not_eq).
rerun(arith : lt_not_sym).
rerun(arith : lt_trans).
rerun(arith : succ_pos).
rerun(arith : pred_neg).
rerun(arith : def_add_succ).
rerun(arith : def_add_pred).
rerun(arith : add_int).
rerun(arith : succ_dist_add_l).
rerun(arith : pred_dist_add_l).
rerun(arith : pred_1_to_1b).
rerun(arith : add_sym).
rerun(arith : le_pred).
save.


add_tactic(induction).
add_tactic(int).


end_of_file.

mkaxiom(succ_dist_add_r, int(A) and int(B) => succ(A + B) = A + succ(B)).
mkaxiom(pred_dist_add_r, int(A) and int(B) => pred(A + B) = A + pred(B)).
mkaxiom(add_assoc, int(A) and (int(B) and int(C)) => A + (B + C) = (A + B) + C).
mkaxiom(eq_add_both_r, int(A) and (int(B) and int(C)) => (A + C = B + C <=> A = B)).
mkaxiom(add_nat, int(A) and (int(B) and (0 =< A and 0 =< B)) => 0 =< A + B).
mkaxiom(nat_induction, (is_local x_1 A) and ([0/x_1]*A and (all x_1 int(x_1) and (0 =< x_1 and A) => [(succ(x_1))/x_1]*A)) => (all x_1 int(x_1) and 0 =< x_1 => A)).
mkaxiom(eq_add_both_l, int(A) and (int(B) and int(C)) => (C + A = C + B <=> A = B)).
mkaxiom(add_inv_l, int(A) => (- A) + A = 0).
mkaxiom(eq_add_move_r, int(A) and (int(B) and int(C)) => (A + C = B <=> A = B + (- C))).
mkaxiom(eq_add_move_l, int(A) and (int(B) and int(C)) => (A = B + C <=> A + (- C) = B)).
mkaxiom(eq_add_neg_zero, int(A) and (int(B) and int(C)) => (A + (- B) = 0 <=> A = B)).
mkaxiom(eq_neg_both, int(A) and int(B) => (- A = - B <=> A = B)).
mkaxiom(eq_add_zero, int(A) and (int(B) and int(C)) => (A + B = 0 <=> A = - B)).
mkaxiom(locallt, (is_local x_1 A) and (is_local x_1 B) => (is_local x_1 A < B)).
mkaxiom(lt_add_both_r, int(A) and (int(B) and int(C)) => (A + C < B + C <=> A < B)).
mkaxiom(lt_add_both_l, int(A) and (int(B) and int(C)) => (C + A < C + B <=> A < B)).
mkaxiom(lt_add_move_r, int(A) and (int(B) and int(C)) => (A + C < B <=> A < B + (- C))).
mkaxiom(lt_add_move_l, int(A) and (int(B) and int(C)) => (A < B + C <=> A + (- C) < B)).
mkaxiom(lt_bool, int(A) and int(B) => bool(A < B)).
mkaxiom(lt_add_neg_zero, int(A) and int(B) => (A + (- B) < 0 <=> A < B)).
mkaxiom(lt_add_zero, int(A) and int(B) => (A + B < 0 <=> A < - B)).
mkaxiom(lt_neg_both, int(A) and int(B) => (- A < - B <=> B < A)).
mkaxiom(lt_neg_r, int(A) => (0 < - A <=> A < 0)).
mkaxiom(lt_neg_l, int(A) => (- A < 0 <=> 0 < A)).
mkaxiom(le_add_both_r, int(A) and (int(B) and int(C)) => (A + C =< B + C <=> A =< B)).
mkaxiom(le_add_both_l, int(A) and (int(B) and int(C)) => (C + A =< C + B <=> A =< B)).
mkaxiom(le_add_move_r, int(A) and (int(B) and int(C)) => (A + C =< B <=> A =< B + (- C))).
mkaxiom(le_add_move_l, int(A) and (int(B) and int(C)) => (A =< B + C <=> A + (- C) =< B)).
mkaxiom(le_add_neg_zero, int(A) and int(B) => (A + (- B) =< 0 <=> A =< B)).
mkaxiom(le_add_zero, int(A) and int(B) => (A + B =< 0 <=> A =< - B)).
mkaxiom(le_neg_both, int(A) and int(B) => (- A =< - B <=> B =< A)).
mkaxiom(le_neg_r, int(A) => (0 =< - A <=> A =< 0)).
mkaxiom(le_neg_l, int(A) => (- A =< 0 <=> 0 =< A)).
mkaxiom(neg_neg_id, int(A) => - (- A) = A).
mkaxiom(mul_int, int(A) and int(B) => int(A * B)).
mkaxiom(mul_zero_r, int(A) => A * 0 = 0).
mkaxiom(mul_zero_l, int(A) => 0 * A = 0).
mkaxiom(mul_dist_succ_l, int(A) and int(B) => succ(A) * B = A * B + B).
mkaxiom(mul_dist_pred_l, int(A) and int(B) => pred(A) * B = A * B + (- B)).
mkaxiom(mul_sym, int(A) and int(B) => A * B = B * A).
mkaxiom(mul_dist_pred_r, int(A) and int(B) => A * pred(B) = A * B + (- A)).
mkaxiom(mul_dist_succ_r, int(A) and int(B) => A * succ(B) = A * B + A).
mkaxiom(mul_dist_add_r, int(A) and (int(B) and int(C)) => A * (B + C) = A * B + A * C).
mkaxiom(mul_dist_add_l, int(A) and (int(B) and int(C)) => (A + B) * C = A * C + B * C).
mkaxiom(neg_zero, -(0) = 0).
mkaxiom(neg_dist_mul_r, int(A) and int(B) => - A * B = A * (- B)).
mkaxiom(mul_assoc, int(A) and (int(B) and int(C)) => A * (B * C) = (A * B) * C).
mkaxiom(neg_dist_mul_l, int(A) and int(B) => - A * B = (- A) * B).
mkaxiom(mul_neg_swap, int(A) and int(B) => (- A) * B = A * (- B)).
mkaxiom(neg_dist_add, int(A) and int(B) => - (A + B) = (- A) + (- B)).
mkaxiom(add3_sym, int(A) and (int(B) and int(C)) => (A + B) + C = (A + C) + B).
mkaxiom(mul3_sym, int(A) and (int(B) and int(C)) => (A * B) * C = (A * C) * B).
mkaxiom(succ_adjacent, int(A) and int(B) => not (A < B and B < succ(A))).
mkaxiom(pred_adjacent, int(A) and int(B) => not (pred(A) < B and B < A)).
mkaxiom(pred_cases, int(A) and (int(B) and (pred(A) =< B and B =< A)) => pred(A) = B or B = A).
mkaxiom(lt_succ_le, int(A) and int(B) => (A < B <=> succ(A) =< B)).

rerun(arith : le_succ_lt).

theorem nat_induction2
=== (is_local x A)
    and [0/x]*A
    and (all x x:nats and A => [(x+1)/x]*A)
    => (all x x:nats => A).

/**************** Postulates about multiplication *****************/
postulate eq_mul_zero
=== int(A) and int(B)
    => (A * B = 0 <=> (A = 0 or B = 0)).

postulate mul_pos_pos
=== int(A) and 0 < A and int(B) and 0 < B
    => 0 < A * B.

postulate mul_nat_nat
=== int(A) and 0 =< A and int(B) and 0 =< B
    => 0 =< A * B.

postulate eq_mul_both_r
=== int(A) and int(B) and int(C) and C \= 0
    => (A * C = B * C <=> A = B).

postulate eq_mul_both_l
=== int(A) and int(B) and int(C) and C \= 0
    => (C * A = C * B <=> A = B).

postulate exp_dist_add
=== int(A) and int(B) and 0 =< B and int(C) and 0 =< C
    => A ** (B + C) = (A ** B) * (A ** C).


/*
 * Integer arithmetic evaluation coded into Qu-Prolog.
 *
 * NB. only positive integer constants should be allowed.
 *      (negative constants are represented as (- K))
 *      However, the rules do not currently check this, so it is
 *      possible (but not recommended) to evaluate negative constants.
 *
 * These rules are currently unsound due to the lack of
 * overflow checking in Qu-Prolog.
 */

axiom eval_int === int(A) :- A1 evals A,
                             integer(A1).
axiom eval_eq  === A = B  :- A1 evals A,
                             B1 evals B,
			     (integer(A1) ; integer(B1)),
			     A1 = B1.
axiom eval_lt  === A < B  :- A is_lt B.
axiom eval_le  === A =< B :- A is_le B.

axiom eval_not_eq === (A = B) <=> false :- A1 evals A, B1 evals B, 
					   integer(A1), integer(B1),
					   \+ (A1 =:= B1).
axiom eval_not_lt === (A < B) <=> false  :- B is_le A.
axiom eval_not_le === (A =< B) <=> false :- B is_lt A.


rerun(arith : succ_eq_add1).
rerun(arith : le_add1_lt).

mkaxiom(lt_add1_le, int(A) and int(B) => (A < B <=> A + 1 =< B)).
mkaxiom(pred_eq_sub1, int(A) => pred(A) = A + (-(1))).
mkaxiom(mul_one_l, int(A) => 1 * A = A).
mkaxiom(mul_one_r, int(A) => A * 1 = A).

rule_group(localness, arith:localadd, r).
rule_group(localness, arith:localneg, r).
rule_group(localness, arith:localint, r).
rule_group(localness, arith:localmul, r).
rule_group(localness, arith:localsucc, r).
rule_group(localness, arith:localpred, r).
rule_group(localness, arith:localexp, r).
rule_group(localness, arith:localle, r).
rule_group(localness, arith:locallt, r).

rule_group(arith_types, add_int, r).
rule_group(arith_types, neg_int, r).
rule_group(arith_types, zero_int, t).
rule_group(arith_types, mul_int, r).
rule_group(arith_types, succ_int, r).
rule_group(arith_types, pred_int, r).
rule_group(arith_types, eval_int, t).
rule_group(arith_types, int_bool, t).
rule_group(arith_types, le_bool, r).
rule_group(arith_types, lt_bool, r).

%% Theorems and tactics for Alena Griffith's linear arithmetic solver.
rerun(arith:add_k_r).
rerun(arith:add_le).
rerun(arith:mul_nat_maint_le).
rerun(arith:linear_comb).
rerun(arith:linear1_comb).

add_tactic(int_simp).
add_tactic(li_prove).

begin_compiled_rule(int(I),      int(I), [precondition=nonvar(I)]).
compiled_rule(r:refleximp,      (int(I) :- var(I),(\+ frozen(I)),!,1=return)).
compiled_rule(r:refleximp,      (int(I) :- 1=is_hyp,!)).
compiled_rule(arith:add_int,    (int(_) :- !,_)).
compiled_rule(arith:neg_int,    (int(_) :- !,_)).
compiled_rule(arith:mul_int,    (int(_) :- !,_)).
compiled_rule(arith:succ_int,   (int(_) :- !,_)).
compiled_rule(arith:pred_int,   (int(_) :- !,_)).
compiled_rule(arith:eval_int,   (int(I) :- integer(I),!,body)).
compiled_rule(arith:nat_to_int, (int(I) :- 1=is_hyp,!)).
compiled_rule(arith:nat_to_int, (int(I) :- atomic(I),1=is_theorem,!)).
compiled_rule(r:refleximp,      (int(I) :- atomic(I),1=is_theorem,!)).

