%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS heur.ql,v 1.13 1993/12/10 02:18:43 marku Exp


:- declare(envir/0).      % prints the current/default environment.
:- declare(envir/3).      % declares a named environment.
:- declare(default_envir/1).  % sets the default environment.
:- declare(heuristic/7).  % declares a heuristic.
:- declare(heuristic/4).  % declares a heuristic.
:- declare(heuristic/3).  % declares a sub-heuristic.

%% heur.ql,v 1.13 1993/12/10 02:18:43 marku Exp

%% default_envir(+EnvirSpec).
%%
%%  This sets the default environment for all proofs that use
%%  the u1 proof subsystem, in the current theory and (possibly)
%%  its descendants.

default_envir(Envir) :-
	envir(Envir),
	is_curr_envir(E),
	current_dir(T),
	retractall(default_heuristic_environment(_)),
	retractall([default_envir, T] : [_]),
	tactic_assert(default_envir, [E]).


%% envir(+Name, +Cutoff, +Levels).
%%
%%	Define a named environment.

envir(Name, Cutoff, Levels) :-
	wff_envir_levels(Levels),
	check_or_writelist(
	    \+  get_up(envir, [Name, _, _]),
	    ['There is already a heuristic environment called: ',Name,nl]),
	tactic_assert(envir, [Name, Cutoff, Levels]).


%%  wff_envir_levels(+Levels).
%%	Checks that Levels is a list of lists and that each
%%	of the sublists only contains known heuristics.

wff_envir_levels([]).

wff_envir_levels([L|Ls]) :-
	wff_heur_list(L),
	wff_envir_levels(Ls).


%%  wff_heur_list(+Hlist).
%%	Checks that all the heuristic names/calls in Hlist
%%	have been properly declared using the heuristic/7 command.

wff_heur_list([]).

wff_heur_list([H|Hs]) :-
	get_up(heuristic,[H,_,_,_,_,_,_]), !,
	wff_heur_list(Hs).

wff_heur_list([H|Hs]) :-
	writelist(['Unknown heuristic: ',H,nl]),
	wff_heur_list(Hs),
	!,
	fail.

wff_heur_list(X) :-
	writelist(['Badly formed list of heuristics: ',X,nl]),
	!,
	fail.




%%
%% heuristic(+HCall, +SubHCall).
%% heuristic(+HCall, +DefLevel, +Doc, +SubHCall).
%% heuristic(+HCall, +DefLevel, +Doc, +Tst, +PreMsg, +Action, +PostMsg).
%%
%% Define multiple - clause heuristics.
%%


heuristic(HCall, SubHCall) :-
	functor(HCall, F, N),
	functor(HSkel, F, N),
	check_or_writelist(
		get_up(heuristic, [HSkel, DefLevel | _]),
		['Heuristic ', F,'/',N,
		 ' improperly defined - use heuristic/4 for first clause.\n']),
	heuristic(HCall, DefLevel, sub_heuristic, SubHCall).

heuristic(HCall, DefLevel, sub_heuristic, SubHCall) :-
        \+ (get_up(heuristic, [SubHCall, _, _, Tst, PreMsg, Action, PostMsg]),
            tactic_assert(heuristic, [HCall, DefLevel, sub_heuristic,
                                          Tst, PreMsg, Action, PostMsg]),
            fail), !.


heuristic(HCall, DefLevel, Doc, SubHCall) :-
	\+ (get_up(heuristic, [SubHCall, _, D2, Tst, PreMsg, Action, PostMsg]),
	    (D2 == sub_heuristic -> D3 = D2 ; D3 = Doc),
   	    tactic_assert(heuristic, [HCall, DefLevel, D3,
                                          Tst, PreMsg, Action, PostMsg]),
	    fail), !.

heuristic(HCall, DefLevel, Doc, SubHCall) :-
	functor(HCall, F, N),
        writelist(['Heuristic ',F,'/',N,' calls unknown sub-heuristic:', 
		   SubHCall, nl]),
	!, fail.




heuristic(HCall, DefLevel, Doc, Tst, PreMsg, Action, PostMsg) :-
	tactic_assert(heuristic, [HCall, DefLevel, Doc,
	                          Tst, PreMsg, Action, PostMsg]).





%% is_curr_envir(-Envir).
%%
%%	Returns the current heuristic environment in Envir.
%%	If no curr_envir is set, it defaults to the one set
%%	by the default_envir/1 command.  If that is not set, it
%%      defaults to the empty environment.

is_curr_envir(Envir) :-
	ipLookup(curr_envir, Envir),
	(
	    (Envir = envir(_,N,L), nonvar(N), nonvar(L))
	;
	    default_heuristic_environment(Envir)
	;
	    once((
	            get_up(default_envir, [Envir])
	        ;
		   (Envir = envir(empty,0,[]))
	        )),
	    asserta(default_heuristic_environment(Envir))
	),
	!.




%% envir/0
%%
%%	Display all the heuristic tables in all levels 

envir :-
	is_curr_envir(envir(Name,Cutoff,Levels)),
	(
	    (Name = current
	    ->
	    true)
	;
	    writelist(['The current environment is environment ''',
                          Name,'''.', nl])
	),
	writelist(['Levels 1..', Cutoff,' are active.',nl]),
	write_envir_levels(Levels, 1, '  ').


write_envir_levels([], _, _).

write_envir_levels([[]|Ls], N, Indent) :-
	!,
	N1 is N + 1,
	write_envir_levels(Ls, N1, Indent).

write_envir_levels([L|Ls], N, Indent) :-
	writelist(Indent),
	writelist([N, ':']),
	\+ (member(H,L),
	    writelist(['  ',H]),
	    fail
	   ),
	nl,
	N1 is N + 1,
	write_envir_levels(Ls, N1, Indent).



%% envir( +EnvirSpec )
%%
%%	Set the current meta-heuristic level and/or select
%%	a named heuristic environment.

envir( E ) :-
	var(E),
	!.   %% Leave the current environment unchanged.

envir( envir(Name, Cutoff, Levels) ) :-
	((Cutoff =:= 0) ; is_valid_level(Cutoff)),
	wff_envir_levels(Levels),
	!,
	ipSet(curr_envir, envir(Name,Cutoff,Levels)).
	
envir( Name ) :-
	atom(Name),
	!,
	get_up(envir, [Name, Cutoff, Levels]),
	ipSet(curr_envir, envir(Name,Cutoff,Levels)).

envir( Name : Cutoff ) :-
	atom(Name),
	is_valid_level(Cutoff),
	!,
	get_up(envir, [Name, _, Levels]),
	ipSet(curr_envir, envir(Name,Cutoff,Levels)).

envir( Cutoff ) :-
	((Cutoff =:= 0) ; is_valid_level(Cutoff)),
	!,
	is_curr_envir(envir(Name,_,Levels)),
	ipSet(curr_envir, envir(Name,Cutoff,Levels)).



%% auto( +EnvirSpec ).
%%
%%	Same as envir/1, but adds an always(quit) heuristic
%%	so that the user is never called.

auto( EnvirSpec ) :-
	envir(EnvirSpec),
	is_curr_envir(envir(_,Cutoff,_)),
	add_heur(always(quit), -Cutoff).



%% add_heur(+H)
%% add_heur(+H,+L)
%%
%%	Adds the heuristic H to level L. 
%%	(The first form sets L to the default level for that heuristic).
%%	It adds it at the beginning if L is positive, otherwise at the end.
%%
%%	These commands change the name of the current environment to 'current'.

add_heur(H) :-
	get_up(heuristic, [H,L,_,_,_,_,_]),
	add_heur(H,L).
	

add_heur(H, N) :-
	N > 0,
	is_valid_level(N),
	get_up(heuristic, [H,_,_,_,_,_,_]),
	!,
	is_curr_envir(envir(_,Cutoff,Levels)),
	append(Levels,[[],[],[],[]],Levels2),
	select(N, LevN, Levels2, [H|LevN], Levels3),
	envir(envir(current,Cutoff,Levels3)).

add_heur(H, Neg) :-
	Neg < 0,
	N is -(Neg),
	is_valid_level(N),
	get_up(heuristic, [H,_,_,_,_,_,_]),
	!,
	is_curr_envir(envir(_,Cutoff,Levels)),
	append(Levels,[[],[],[],[]],Levels2),
	select(N, LevN, Levels, LevN2, Levels2),
	append(LevN, [H], LevN2),
	envir(envir(current,Cutoff,Levels2)).



%% delete_heur(+Patt)
%% delete_heur(+Patt,+N)
%%
%%	Removes the first heuristic in the current environment
%%	(at level N only, if N is specified) that matches Patt.
%%
%%	These commands change the name of the current environment to 'current'.

delete_heur(Patt) :-
	delete_heur(Patt,_).

delete_heur(Patt,N) :-
	is_curr_envir(envir(_,Cutoff,Levels)),
	is_valid_level(N),
	select(N, LevN, Levels, LevN2, Levels2),
	append(Front, [Patt|Rear], LevN),
	!,
	append(Front, Rear, LevN2),
	envir(envir(current,Cutoff,Levels2)).
	


%% envir_level(+Num, +HeurList).
%%
%%	Replaces level Num of the current heuristic environment
%%	with the given HeurList.
%%
%%	This command changes the name of the current environment to 'current'.

envir_level(N, Heurs) :-
	is_valid_level(N),
	wff_heur_list(Heurs),
	is_curr_envir(envir(_,Cutoff,Levels)),
	select(N, LevN, Levels, Heurs, Levels2),
	envir(envir(current,Cutoff,Levels2)),
	!.
	


%% name_curr_envir(+Name).
%%
%%	Copies the current environment and gives it a name.
%%      It also changes the current environment to be called Name.

name_curr_envir(Name) :-
	atom(Name),
	is_curr_envir(envir(_,Cutoff,Levels)),
	tactic_assert(envir,[Name,Cutoff,Levels]),
	envir(envir(Name,Cutoff,Levels)).



%% try_heuristics(+Action).
%%
%%	This is called before each window is printed out, so that
%%	active heuristics in the current environment can be tried.
%%
%%      It returns the Action that it executed, so the caller can
%%      terminate if Action==quit.  A better way of handling this
%%      might be to write a quit tactic that does ipSet(quit_flag,true)
%%      and change call_tac to recognise this flag and exit?

try_heuristics(Action) :-
	is_curr_envir(envir(_, Cutoff, Levels)),
	is_valid_level(N),
	N =< Cutoff,
	nth1(N, Levels, LevN),
	member(H, LevN),
	call_heuristic(H, Action).


call_heuristic(H) :- 
	call_heuristic(H, _).

call_heuristic(H, Action) :-
	get_up(heuristic, [H, _, _, Tst, PreMsg, Action, PostMsg]),
	Tst,
	msg(PreMsg),
	(
	    (Action == quit)
	;
	    once(Action)
	),
	!,
	msg(PostMsg).



%% hints(?Cutoff, -HeurList).
%%
%%	This returns a list of all the heuristics in the current 
%%	environment whose tests succeed on the current window.  
%%	Each element of the resulting HeurList has the form:
%%
%%	     heuristic(Call, Level, Doc, Tst, PreMsg, Action, PostMsg)
%%
%%	Note that Tst has already been executed successfully and may
%%	have instantiated the other components.
%%
%%	If Cutoff is a variable, all levels of the current environment
%%	are searched, otherwise it must be a valid level number and
%%	only levels up to and including Cutoff are searched.

hints(N, Hlist) :-
	var(N),
	is_curr_envir(envir(_, _, Levels)),
	hints_find(Levels,1,4,Hlist).

hints(N, Hlist) :-
	is_valid_level(N),
	is_curr_envir(envir(_, _, Levels)),
	hints_find(Levels,1,N,Hlist).


%%  hints_find(+Levels, +LevNum, +Cutoff, -HintList).

hints_find([L|Ls], N, Cutoff, Hlist) :-
	N =< Cutoff,
	!,
	hints_find_in_level(L, N, HN),
	append(HN, HRest, Hlist),
	N1 is N + 1,
	hints_find(Ls, N1, Cutoff, HRest).

hints_find(_, _, _, []).


%%  hints_find_in_level(+Level, +LevNum, -HintList).

hints_find_in_level([], _, []).

hints_find_in_level([H|Hs], L, [heuristic(H,L,Doc,Tst,Pre,Act,Post)|Rest]) :-
	get_up(heuristic, [H,_,Doc,Tst,Pre,Act,Post]),
	once(Tst),
	!,
	hints_find_in_level(Hs, L, Rest).

hints_find_in_level([_|Hs], L, Rest) :-
	hints_find_in_level(Hs, L, Rest).



%% hints.
%% hints(+Cutoff).
%%
%%	These print a menu of all heuristics in the current
%%	environment whose tests succeed on the current window.
%%
%%	If Cutoff is a variable, all levels of the current environment
%%	are searched, otherwise it must be a valid level number and
%%	only levels up to and including Cutoff are searched.


hints :-
	hints(_).

hints(Cutoff) :-
	once(hints(Cutoff, Hints)),
	(
	    (Hints = []
	    ->
	    writelist(['No applicable hints.',nl]))
	;
	    true
	),
	member(heuristic(H,L,_,_,Pre,_,_), Hints),
	writelist(['  ',L,': ',H,nl,'        ']),
	writelist(Pre),
	nl,
	fail.

hints(_).




%% is_valid_level(?Level)
%%
%%	Check that Level is in the range 1..4, or generate it.


is_valid_level(Level) :-
	nonvar(Level), !,
	check_or_writelist(
		member(Level, [1,2,3,4]),
		['Error: illegal level number: ', Level]).

is_valid_level(Level) :-
	member(Level, [1,2,3,4]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%   Auxillary prediates that probably are/should be in some library.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% nth1(+N, ?List, ?Val).
%%
%%	Selects the 'N'th element of List and unifies it with Val.
%%      Note that the numbers start from 1.

nth1(1, [V | _], X) :-
	!,
	X = V.

nth1(N, [_ | Ls], V) :-
	N > 1,
	N1 is N - 1,
	nth1(N1, Ls, V).


%% select(+N, ?X0, ?List0, ?X, ?List)
%%
%%	From Ray Nickson's lists.pl

select(1, X, [X|L], Y, [Y|L]).

select(N, X, [A|XL], Y, [A|YL]) :-
    N > 1,
    N1 is N - 1,
    select(N1, X, XL, Y, YL).
