%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS li_prove.ql,v 1.6 1993/12/10 02:18:52 marku Exp


/*********************************************************
**********************************************************

	A TACTIC TO PROVE A LINEAR INEQUALITY
	=====================================


This tactic takes a list of hypotheses in
standard form, reads them into a list structure
suitable for input to the Fourier-Motkin algorithm,
then analyzes them for consistency.

If an inconsistency is detected, the user is 
alerted and the necessary proof
steps are carried out, resulting eventually in
a false hypothesis.  */


:- tactic(li_prove(Hyp_list),
	['li_prove(HypList) --- use linear arithmetic to deduce false.',nl,
	 'HypList must be a list of hypothesis numbers.',nl,
	 'This tactic attempts to find a contradiction between those',nl,
	 'hypotheses using the Fourier-Motzkin linear arithmetic algorithm'],
	'using Fourier-Motzkin algorithm',
	[]).

%% Temporary: environment for the li_prove tactic...
	
:- envir(li_proofs, 4,
	[ [use_hyp,quit],
	  [conjuncts, discharge_hyp],
	  [when_focus(int(_), use(arith:int, true)),
	   when((is_focus(A =< B), A > B),
	        use(arith:eval_not_le,false)),
	   when((is_focus(A =< B), A =< B),
	        use(arith:eval_le,true))]
	]).


/**********************************************************
***********************************************************
***********************************************************

		TOP LEVEL VIEW
		==============

The tactic consists of two distinct phases.
First, the tactic "thinks ahead" to decide whether
a contradiction can be deduced. If it decides that
this is possible, a record of necessary proof steps
is returned, and the tactic then proceeds with
phase 2, and carries out the necessary proof steps
to arrive at the contradiction. 

li_prove(+Hyp_list)
*/
li_prove(Hyp_list):-	
	test_fm(Hyp_list,List_lists,Vars,Lambdas,Bad_rhs), 
	nl,
	write_string("Contradiction in hypotheses detected - proceeding with Fourier-Motzkin reduction"),
	nl,
	red_fm(Hyp_list, Bad_rhs).


/*********************************************************
**********************************************************

PHASE 1 - TESTING COMPONENT 
===========================

In this phase of the tactic the hypothesis list is analyzed
for consistency. The goal "test_fm" succeeds only if an
inconsistency is detected. Two distinct steps 
are involved:-
	i) first, the hypotheses are "processed" into  
	data structures suitable for speedy analysis by
	the 'consistent' routine. These data structures
	are used internally, and are also output to
	Phase 2 of the tactic.

	ii) next, the 'consistent' routine analyzes the
	data structures, searching for inconsistencies.

test_fm(+Hyp_list, -List_lists, -Vars, -Lambdas, -Bad_rhs)
*/

test_fm(Hyp_list, List_lists, Vars, Lambdas, Bad_rhs):-
	read_in(Hyp_list, List_lists, Vars, Lambdas),
	!,
	consistent(List_lists, Vars, Lambdas, not_sat,
		Bad_rhs).

/********************************************************

PH(1) - part i)

Processing the hypotheses.

read_in(+Hyp_list, -List_lists, -Vars, -Lambdas)
*/


/* **************************************************
	COMMENCEMENT OF CODE FOR read_in
 **************************************************
co_eff_val(+Term, +Side, -Value).

Returns the value of a co-efficient. The sign will
be affected according to whether the co-efficient
is on the lhs, or rhs of the equation.

*/

co_eff_val(-(T),rhs,T).

co_eff_val(-(T),lhs,C):-
	C is -T.

co_eff_val(T, rhs, C):-
	C is -T.

co_eff_val(T, lhs, T).

/* **************************************************

	AUXILIARY FUNCTIONS

mem(+Elem, +List): Elem is a member of List.
get_pos(+Elem, +List, -Pos): Pos is the position of
	Elem in List.
increment(+L, +Pos, +Val, -New_l): New_l is obtained
	by incrementing the Pos'th element in 
	L by Val.
*/

mem(E, [H|Tail]):-
	or(E == H, mem(E, Tail)).

get_pos(X,[H|_],1):-
	X == H.

get_pos(X,[H|Tail], Pos):-
	get_pos(X,Tail,New_pos),
	Pos is New_pos + 1.

get_pos_of_1([1|_],1).

get_pos_of_1([H|Tail],Pos):-
	get_pos_of_1(Tail,N),
	Pos is N + 1.

increment([H|Tail],1,Val,New_list):-
	Inc_val is Val + H,
	New_list = [Inc_val|Tail].

increment([H|Tail],Pos,Val,New_list):-        
	New_pos is Pos - 1,
	increment(Tail, New_pos, Val, New_l),
	New_list = [H|New_l].

/* **************************************************
augment(+Term,+Side,+Next_ineq,+Vars_in,+Ls_in,
		-New_ineq, -Vars_out, -Ls_out)

	Term ::= X =< Y | X+Y | X*Y | -(X) | X
	Side ::= none | lhs | rhs
	Next_ineq ::= [ [...], [...] ]
	Vars_in ::= list of vars in the system
	Ls_in ::= list of constants on rhs of system

*/

augment(X =< Y, none, Ineq, V_in, L_in,
		New_ineq, V_out, L_out):-
	augment(X, lhs, Ineq, V_in, L_in,
			Ineq_a, V_out_a, L_out_a),
	augment(Y, rhs, Ineq_a, V_out_a, L_out_a,
			New_ineq, V_out, L_out).

augment(X + Y, Side, Ineq, V_in, L_in,
		New_ineq, V_out, L_out):-
	augment(X, Side, Ineq, V_in, L_in,
			Ineq_a, V_out_a, L_out_a),
	augment(Y, Side, Ineq_a, V_out_a, L_out_a,
			New_ineq, V_out, L_out).



augment(X*Y, Side, [L,R], V_in, L_in,
		New_ineq, V_in, L_in):-
	mem(Y, V_in),
	get_pos(Y, V_in, Pos),
	co_eff_val(X, Side, Val),
	increment(L, Pos, Val, New_l),
	New_ineq = [New_l,R].

augment(X*Y, Side, [L,R], V_in, L_in,
		New_ineq, V_out, L_in):-
	not(mem(Y, V_in)),
	length(V_in,Nos_var),
	append(V_in,[Y],V_out),
	co_eff_val(X, Side, Val),
	append(L, [Val], New_l),
	New_ineq = [New_l,R].

augment(-(T), lhs, [L,R], V_in, L_in,
		Ineq, V_in, L_out):-
	get_pos_of_1(R, Inequality_nos),
	increment(L_in, Inequality_nos, T, L_out),
	Ineq = [L,R].

augment(-(T), rhs, [L,R], V_in, L_in,
		Ineq, V_in, L_out):-
	get_pos_of_1(R, Inequality_nos),
	Val is -T,
	increment(L_in, Inequality_nos, Val, L_out),
	Ineq = [L,R].

augment(T, lhs, [L,R], V_in, L_in,
		Ineq, V_in, L_out):-
	get_pos_of_1(R, Inequality_nos),
	Val is -T,
	increment(L_in, Inequality_nos, Val, L_out),
	Ineq = [L,R].

augment(T, rhs, [L,R], V_in, L_in,       
		Ineq, V_in, L_out):-
	get_pos_of_1(R, Inequality_nos),
	increment(L_in, Inequality_nos, T, L_out),
	Ineq = [L,R].


/* ********************************************************

make_sys(+Nos_equations,-New_sys)
	This routine makes a system of the form
		[ [ [], [1,0,0, ..,0] ],
		  [ [], [0,1,0, ..,0] ],
		  [ [], [0,0,1, ..,0] ],
		       ...
		  [ [], [0,0,0, ..,1] ] ].
where the number of lists in the outermost list
is Nos_equations.

*/

make_sys(Nos_eqs, New_sys):-
	add_eq([], 1, Nos_eqs, New_sys).

add_eq(Sys_list, Nos_eqs, Nos_eqs, New_sys):-
	!,
	make_rhs(Nos_eqs, Nos_eqs, Last_eq),
	append(Sys_list,[[ [], Last_eq ]], New_sys).

add_eq(Sys_list, Eq_nos, Nos_eqs, New_sys):-
	make_rhs(Eq_nos, Nos_eqs, New_eq),
	append(Sys_list, [[ [], New_eq ]], Int_sys),
	New_eq_nos is Eq_nos + 1,
	add_eq(Int_sys, New_eq_nos, Nos_eqs, New_sys).

make_rhs(Eq_nos, Nos_eqs, New_rhs):-
	make_list_zeroes(Nos_eqs, Z_list),
	increment(Z_list, Eq_nos, 1, New_rhs).

make_list_zeroes(1,[0]) :-
	!.

make_list_zeroes(X,L):-
	Y is X - 1,
	make_list_zeroes(Y,New_l),
	append(New_l, [0], L).


/* ***************************************************

make_same_length(+L1, +L2, -L3).
	Makes L1 the same length as L2 by appending
	zeroes to the end of L1 to get L3.


*/

make_same_length(L1, L2, L1):-
	length(L1, X),
	length(L2, X).

make_same_length(L1, L2, L3):-
	length(L1, X),
	length(L2, Y),
	X < Y,
	append(L1, [0], L),
	make_same_length(L, L2, L3).

equalize_lhs_length([],L,[]).

equalize_lhs_length([[Lhs,Rhs]|Rest],L,Sys_out):-
	make_same_length(Lhs, L, Lhs_n),
	equalize_lhs_length(Rest, L, Sys),
	Sys_out = [[Lhs_n,Rhs]|Sys].


/* **************************************************

add_in(+Hyp_nos_list, +Sys_in, +Vars_in, +Lambdas_in,
	-Sys_out, -Vars_out, -Lambdas_out).

This routine takes each hyp. nos from +Hyp_nos_list
in turn, and gets the statement of the hypothesis.
It must be in the form,

	S ::= Side =< Side
	Side ::= Side + Side | Side*Side | -(Side) | nos

The stmt is used to augment +Sys_in, +Vars_in, and
+Lambdas_in, to get -Sys_out, -Vars_out, and
-Lambdas_out respectively.
*/

add_in([], [], Vars, Lambdas,
		[], Vars, Lambdas).


add_in([Hyp_nos|Rest_hyps],[New_ineq|Rest_sys], Vars_in,
	Lambdas_in,
		New_sys, Vars_out, Lambdas_out):-
	is_hyp(Hyp_nos, Term),
	New_ineq = [Lhs, Rhs],
	make_same_length(Lhs, Vars_in, Lhs1),
	New_ineq_1 = [Lhs1, Rhs],
	augment(Term, none, New_ineq_1, Vars_in, 
		Lambdas_in, New_ineq_a, Vars_i, Ls_i),
	add_in(Rest_hyps, Rest_sys, Vars_i, Ls_i,
		Int_sys, Vars_out, Lambdas_out),
	Int_sys_1 = [New_ineq_a|Int_sys],
	equalize_lhs_length(Int_sys_1, Vars_out, New_sys).

/* **************************************************

read_in(+Hyp_list, -System, -Vars, -Lambdas)

This routine reads in the hypotheses whose nos.
are in +Hyp_list, and produces -System, -Vars, and
-Lambdas, suitable for input to the Fourier-Motzkin
algorithm.

*/

read_in(Hyp_list, System, Vars, Lambdas):-
	write_string("Performing the read in procedure"),
	nl,
	length(Hyp_list, Nos_ineqs),
	once(make_sys(Nos_ineqs, New_sys)),
	once(make_list_zeroes(Nos_ineqs, Lambdas_in)),
	add_in(Hyp_list, New_sys, [], Lambdas_in,
		System, Vars, Lambdas).


/********************************************************
	CONCLUSION OF CODE FOR read_in
******************************************************

PH(1) - part ii)

Performing the analysis.

This phase uses the data structure returned from the 
read_in phase and performs a Fourier-Motzkin analysis
technique, searching for inconsistencies. It stops
as soon as one is found, returning the "recipe" (ie.
combination of inequalities necessary) for a contradiction.
If the system is consistent, then that information
is returned also.


consistent(+List_lists, +Vars, +Lambdas, -Sat, -Bad_rhs)

*/

/* *****************************************************
	COMMENCEMENT OF CODE FOR consistent
*******************************************************
*/

write_list([]):-
	nl.

write_list([Head|Tail]):-
	write(Head), tab(1), write_list(Tail).

write_string([]).

write_string([Head|Tail]):-
	put(Head), write_string(Tail).


/*********************************************************************

write_ineq_with_rhs(+Co_eff, +Vars, +R_co_eff, +Lambda)

*/

write_ineq_with_rhs(Co_eff, Vars, R_co_eff, Lambda):-
	write_left(Co_eff, Vars),
	write_right_comp(R_co_eff, Lambda).

write_ineq(Co_eff, Vars, Rhs):-
	write_left(Co_eff, Vars),
	write_right_simp(Rhs).




/*********************************************************************

write_left(+Co_eff, +Vars)
*/

write_left([Headc|[]], [Headv|[]]):-
	write(Headc),
	write(Headv),
	tab(2),
	write_string("=<"),
	tab(2).

write_left([Headc|Tailc], [Headv|Tailv]):-
	write(Headc),
	write(Headv),
	tab(2),
	write_string("+"),
	tab(2),
	write_left(Tailc, Tailv).


/*********************************************************************

write_right_comp(+Co_eff, +Lambda)

*/

write_right_comp(Co_eff, Lambda):-
	dot_prod(Co_eff, Lambda, Res),
	write(Res),
	nl.

write_right_simp([L]):-
	write(L),
	nl.



/*********************************************************************

write_sys_right(List_ineq, Vars, Lambda):-

*/

write_sys_right([], _, _).

write_sys_right([[Lhs1, Rhs1]|Rest], Vars, Lambda):-
	write_ineq_with_rhs(Lhs1, Vars, Rhs1, Lambda),
	write_sys_right(Rest, Vars, Lambda).

write_sys([], _).

write_sys([[Lhs1, Rhs1],[Lhs2,Rhs2]], Vars):-
	write_ineq(Lhs1,  Vars, Rhs1),
	write_ineq(Lhs2, Vars, Rhs2).


dot_prod([], [], 0).

dot_prod([Head1|Tail1], [Head2|Tail2], Res):-
	Inc is Head1*Head2,
	dot_prod(Tail1, Tail2, New_res),
	Res is New_res + Inc.

scal_prod(_, [], []).

scal_prod(S, [Head|Tail], L):-
	scal_prod(S, Tail, L1),
	E is Head*S,
	L = [E|L1].

/* ***************************************

simp_rhs(+List_ineqs, +Lambda, -Simp_ineq).



*/

simp_rhs([], _, []).

simp_rhs([[Head|Tail]|Rest], Lambda, Simp_ineq):-
	Tail = [Rhs],	
	dot_prod(Rhs, Lambda, Res),
	New_bit = [[Head|[[Res]]]],
	simp_rhs(Rest, Lambda, Simp_list),
	append(New_bit, Simp_list, Simp_ineq).




/* ****************************************

amalg_pos(+lhs1, +rhs1, +lhs2, +rhs2, +Lams, -Res)

amalg_neg(+lhs1, +rhs1, +lhs2, +rhs2, +Lams, -Res)

*/

amalg_pos(L1, R1, L2, R2, Lams, Res):-
	New_l is L1*L2,
	scal_prod(L2, R1, New_r1),
	scal_prod(L1, R2, New_r2),
	dot_prod(New_r1, Lams, N1),
	dot_prod(New_r2, Lams, N2),
	N2 =< N1,
	Res = [[New_l],New_r2].


amalg_pos(L1, R1, L2, R2, Lams, Res):-
        New_l is L1*L2,
        scal_prod(L2, R1, New_r1),
        scal_prod(L1, R2, New_r2),
        dot_prod(New_r1, Lams, N1),
        dot_prod(New_r2, Lams, N2),
        N2 > N1,
        Res = [[New_l],New_r1].
 

amalg_neg(L1, R1, L2, R2, Lams, Res):-
        New_l is L1*(-L2),
        L2_m is -L2,
	L1_m is -L1,
	scal_prod(L2_m, R1, New_r1),
	scal_prod(L1_m, R2, New_r2),
        dot_prod(New_r1, Lams, N1),
        dot_prod(New_r2, Lams, N2),
	N2 =< N1,
        Res = [[New_l],New_r2].


amalg_neg(L1, R1, L2, R2, Lams, Res):-
        New_l is L1*(-L2),
        L2_m is -L2,
        L1_m is -L1,
        scal_prod(L2_m, R1, New_r1),
        scal_prod(L1_m, R2, New_r2),
        dot_prod(New_r1, Lams, N1),
        dot_prod(New_r2, Lams, N2),
        N2 > N1,
        Res = [[New_l],New_r1].


/*******************************************

amalg(+Inequality1, +ListInequalities, +Lambdas, -NewList).

NewList contains no more than one inequality with
 positive coeff. and no more than one inequality with a 
negative coeff.

*/

amalg(L1, [], _, [L1]).

amalg(L1, [L2|[]], _, [L1|[L2]]):-
	L1 = [[Lhs1],Rhs1],
	L2 = [[Lhs2],Rhs2],
	or((Lhs1 > 0, Lhs2 < 0),(Lhs1 < 0, Lhs2 > 0)).


amalg([Head1,Tail1], [[Head2,Tail2]|Rest], Lams, Output_list):-
	Head1 = [Lhs1],
	Head2 = [Lhs2],
	or((Lhs1 > 0, Lhs2 < 0),(Lhs1 < 0, Lhs2 > 0)),
	amalg([Head1,Tail1], Rest, Lams, First_out),
	append([[Head2,Tail2]], First_out, Second_out),
	Second_out = [Head3|Tail3],
	amalg(Head3, Tail3, Lams, Output_list).

amalg([Head1,Tail1], [[Head2,Tail2]|Rest], Lams, Output_list):-
	Head1 = [Lhs1],
	Head2 = [Lhs2],
	Lhs1 > 0,
	Lhs2 > 0,
	amalg_pos(Lhs1, Tail1, Lhs2, Tail2, Lams, Res),
	amalg(Res, Rest, Lams, Output_list).

amalg([Head1,Tail1], [[Head2,Tail2]|Rest], Lams, Output_list):-
        Head1 = [Lhs1],
        Head2 = [Lhs2],
        Lhs1 < 0,
        Lhs2 < 0,
        amalg_neg(Lhs1, Tail1, Lhs2, Tail2, Lams, Res),
        amalg(Res, Rest, Lams, Output_list).



/***************************************************/


sat_a([Head|[]], Lams, can_be_sat).

sat_a([[[Lhs1],Rhs1],[[Lhs2],Rhs2]], Lams, Sat):-
	Lhs1 < 0,
	scal_prod(Lhs2, Rhs1, New_right1), 
	Lhs1_m is -Lhs1,
	scal_prod(Lhs1_m, Rhs2, New_right2), 
	dot_prod(New_right1, Lams, N1),
	dot_prod(New_right2, Lams, N2),
	(-N1) =< N2,
	Sat = can_be_sat.

sat_a([[[Lhs1],Rhs1],[[Lhs2],Rhs2]], Lams, Sat):-
	Lhs1 > 0,
	Lhs2_m is -Lhs2,
	scal_prod(Lhs2_m, Rhs1, New_right1),
	scal_prod(Lhs1, Rhs2, New_right2), 
	dot_prod(New_right1, Lams, N1),
        dot_prod(New_right2, Lams, N2),
        (-N2) =< N1,
	Sat = can_be_sat.

sat_a(_, _, not_sat).



/********************************************************/

satisfiable(List_inequalities,Lambda_vector,Var_left,Sat,Bad_rhs):-
	write_string("Is this system satisfiable?"),nl,
	write_string("==========================="),nl,
	nl,
	List_inequalities = [Head|Tail],
	amalg(Head, Tail, Lambda_vector, List3),
	!,
	write_string("System reduces to:"),nl,
	write_sys_right(List3, [Var_left], Lambda_vector),
	!,
	sat_a(List3, Lambda_vector, Sat),
	!,
	if( Sat = not_sat,
		 (List3 = [ [ [C1],Rhs1 ], [ [C2],Rhs2 ] ],
		 if(C1 < 0,
			(NewC1 is -C1,
			 NewC2 = C2),
			(NewC2 is -C2,
			 NewC1 = C1)),
		 linear(Rhs1, NewC2, Rhs2, NewC1, Bad_rhs))),
	!.




/* *****************************************************

consistent(+Input_sys, +Vars, +Lambdas, -Sat, -Bad_rhs)

If there are solutions to satisfy the system, then
Sat = sat, otherwise Sat = not_sat.

If Sat = not_sat, then Bad_rhs contains a record of the 
combination of the original hypotheses necessary to 
obtain the contradiction.

*/

consistent(Input_sys, Vars, Lambdas, Sat, Bad_rhs):-
	fm(Input_sys, Vars, Lambdas, Inconsistency, Red_sys, Var_left, Bad_rhs),
	!,
	if(Inconsistency = false, /* OK so far! */
		(satisfiable(Red_sys, Lambdas, Var_left, Sat, Bad_rhs),
		 write_string("This system is "),
		 write(Sat),
		 nl),
		(Sat = not_sat,
		 write_string("This system is not_sat"),
		 nl)),
	!.

/* 
elem_in_pos
===========
This predicate returns the element of the list
that is in Position position.
*/

elem_in_pos([Head|Tail], 1, Co_eff):-
	Co_eff = Head.

elem_in_pos([Head|Tail], Position, Co_eff):-
	Position > 1,
	New_pos is (Position - 1),
	elem_in_pos(Tail, New_pos, Co_eff).


extract_elem([Head|Tail], 1, Result):-
	Result = Tail.

extract_elem([Head|Tail], Pos, Result):-
	Pos > 1,
	New_pos is (Pos - 1),
	extract_elem(Tail, New_pos, L),
	Result = [Head|L].


nos_non_zero_elems([], 0).

nos_non_zero_elems([Head|Tail], Nz):-
	Head = 0,
	nos_non_zero_elems(Tail, N),
	Nz is N.

nos_non_zero_elems([Head|Tail], Nz):-
	not(Head = 0),
        nos_non_zero_elems(Tail, N),
        Nz is (N + 1).



/*
if
==
This predicate evaluates ...
*/

if(Test, Code):-
	Test,
	!,
	Code.
if(Test, Code):-
	true.

if(Test, Code_then, Code_else):-
	Test,
	!,
	Code_then.
if(Test, Code_then, Code_else):-
	Code_else.

or(Code1, Code2):-
	Code1.

or(Code1, Code2):-
	Code2.

not(X):-
	X,!,fail.

not(X).

length_list([],0).

length_list([Head|Tail], X):-
	length_list(Tail, Y),
	X is (Y + 1).


get_signs([],_,0,0,0).

get_signs([ [First_list|_] | Rest_lists], Position,
		Nos_neg, Nos_pos, Nos_zero):-
        elem_in_pos(First_list, Position, Co_eff),
        Co_eff < 0,
        get_signs(Rest_lists, Position, Neg,
		Nos_pos, Nos_zero),
	Nos_neg is Neg + 1.

get_signs([ [First_list|_] | Rest_lists], Position,
                Nos_neg, Nos_pos, Nos_zero):-
        elem_in_pos(First_list, Position, Co_eff),
        Co_eff = 0,
	get_signs(Rest_lists, Position, Nos_neg,
		Nos_pos, Zero),
	Nos_zero is Zero + 1.


get_signs([ [First_list|_] | Rest_lists], Position,
                Nos_neg, Nos_pos, Nos_zero):-
        elem_in_pos(First_list, Position, Co_eff),
        Co_eff > 0,
	get_signs(Rest_lists, Position, Nos_neg,
		Pos, Nos_zero),
	Nos_pos is Pos + 1.


calc_eff(List_lists, Position, Eff):-
        get_signs(List_lists, Position, 
		Nos_neg, Nos_pos, Nos_zero),
        Eff is Nos_pos*Nos_neg + Nos_zero.


nos_var_left([ [Head|_] | _ ], X):-
	length_list(Head, X).

/********************************************************

calc_best(+List_ineq, +Best_pos_so_far, +Best_eff_so_far, +Current_pos, 
		-Best_pos, -Best_eff):-

*********************************************************/

calc_best(_, Best_pos_so_far, Best_eff_so_far, 0, 
		Best_pos_so_far, Best_eff_so_far).

calc_best(List_ineq, Best_pos_so_far, Best_eff_so_far, Curr_pos,
		Best_pos, Best_eff):-
	calc_eff(List_ineq, Curr_pos, Eff),
	Eff >= Best_eff_so_far,
	New_pos is Curr_pos - 1,
	calc_best(List_ineq, Best_pos_so_far, Best_eff_so_far, New_pos,
		Best_pos, Best_eff).

calc_best(List_ineq, Best_pos_so_far, Best_eff_so_far, Curr_pos,
		Best_pos, Best_eff):-
        calc_eff(List_ineq, Curr_pos, Eff),
	Eff < Best_eff_so_far,
	New_pos is Curr_pos - 1,
	calc_best(List_ineq, Curr_pos, Eff, New_pos,
		Best_pos, Best_eff).


/* ************************************************

best_elim(+List_inequalities, -Pos_of_best)
=========

This routine takes as input a list of inequalities
and computes the best variable to eliminate,
as determined by calculation of the "elimination
factor" of each variable.
*/


best_elim(List_inequalities, Pos_of_best) :-
	nos_var_left(List_inequalities, Nos_var),
	calc_eff(List_inequalities, Nos_var, Best_eff_so_far),
	Nos_left is Nos_var - 1,
	calc_best(List_inequalities, Nos_var, Best_eff_so_far, Nos_left,
		Pos_of_best, Best_eff).


/* OK below */

linear([],_,[],_,[]).

linear([H1|T1], F1, [H2|T2], F2, [Hr|Tr]):-
	Hr is H1*F1 + H2*F2,
	linear(T1, F1, T2, F2, Tr).


lin_combine([], _, [], _, []).

lin_combine([H1|T1], F1, [H2|T2], F2, [Hr|Tr]):-
	linear(H1, F1, H2, F2, Hr),
	lin_combine(T1, F1, T2, F2, Tr).


pairwise_elim([Head|Tail], [], Pos, []):-
	elem_in_pos(Head, Pos, Elem),
	not(Elem = 0).

pairwise_elim([Head|Tail], [], Pos, Output_list):-
	elem_in_pos(Head, Pos, Elem),
	Elem = 0,
	extract_elem(Head, Pos, New_bit),
	Output_list = [[New_bit|Tail]].

pairwise_elim([I_head|I_tail], [[J_head|J_tail]|Tail], 
		Pos, Output_list):-
	elem_in_pos(I_head, Pos, C1),
	elem_in_pos(J_head, Pos, C2),
	C1 =< 0, 
	C2 =< 0,
	pairwise_elim([I_head|I_tail], Tail, Pos, Output_list).


pairwise_elim([I_head|I_tail], [[J_head|J_tail]|Tail], 
                Pos, Output_list):-
        elem_in_pos(I_head, Pos, C1),
        elem_in_pos(J_head, Pos, C2),
        C1 >= 0, 
        C2 >= 0,
        pairwise_elim([I_head|I_tail], Tail, Pos, Output_list
).


pairwise_elim([I_head|I_tail], [[J_head|J_tail]|Tail], 
		Pos, Output_list):-
        elem_in_pos(I_head, Pos, C1),
        elem_in_pos(J_head, Pos, C2),
        C1 > 0, 
	C2 < 0,
        New_c2 is -C2,
	lin_combine([I_head|I_tail], New_c2, [J_head|J_tail],
		C1, [N_head|N_tail]),
        extract_elem(N_head, Pos, New_bit),
        pairwise_elim([I_head|I_tail], Tail, Pos, New_list),
	Output_list = [[New_bit|N_tail]|New_list].


pairwise_elim([I_head|I_tail], [[J_head|J_tail]|Tail], 
                Pos, Output_list):-
        elem_in_pos(I_head, Pos, C1),
        elem_in_pos(J_head, Pos, C2),
        C1 < 0, 
        C2 > 0,
	New_c1 is -C1,
	lin_combine([I_head|I_tail], C2, [J_head|J_tail],
                New_c1, [N_head|N_tail]),
        extract_elem(N_head, Pos, New_bit),
        pairwise_elim([I_head|I_tail], Tail, Pos, New_list),
        Output_list = [[New_bit|N_tail]|New_list].




recursively_elim([], Pos, []).


recursively_elim([Head|Tail], Pos, Output_list):-
	pairwise_elim(Head, Tail, Pos, New_ineqs),
	recursively_elim(Tail, Pos, Other_new_ineqs),
	append(New_ineqs, Other_new_ineqs, Output_list).

elim(Input_list, Pos, Output_list):-
	recursively_elim(Input_list, Pos, Output_list).

/* OK below */

eliminate(Input_inequalities, Solution_vector,
  Eliminant, Elim_list, New_vector):-
	best_elim(Input_inequalities, Var_to_elim),
	elem_in_pos(Solution_vector, Var_to_elim, X),
	extract_elem(Solution_vector, Var_to_elim, New_vector),
	Eliminant = [X],
	elim(Input_inequalities, Var_to_elim, Elim_list).






type_1_red(_, [], Out_list):-
	Out_list = [].


type_1_red(T, [[Head|Tail]|Rest_in], Out_list):-
	nos_non_zero_elems(Tail, Nz),
	Nz > T + 1,
	type_1_red(T, Rest_in, New_out),
	Out_list = New_out.


type_1_red(T, [[Head|Tail]|Rest_in], Out_list):-
	nos_non_zero_elems(Tail, Nz),
	Nz =< T + 1,
	Front_bit = [Head|Tail],
	type_1_red(T, Rest_in, New_out),
	Out_list = [Front_bit|New_out].



/* ********************************************************

type_2_red(+Input_sys, +Lambdas, 
	-Output_sys, -Failure, -Bad_rhs)

This routine takes a list of inequalities and checks to 
see whether the lhs of any of them evaluates to zero.

If so, then if the rhs evaluates to N, where N >= 0, the
the inequality is redundant. If N < 0, then the inequality
is unsatisfiable, so we are finished. In this case, Failure
is set to true, and the rhs of the offending inequality is
returned.

*/

type_2_red([],_,[],false,_).


type_2_red([[Lhs1, Rhs1]|Tail],Lambdas,
		Out_list,Failure,Bad_rhs):-
	not(nos_non_zero_elems(Lhs1,0)),
	type_2_red(Tail,Lambdas,New_out,Failure,Bad_rhs),
	Out_list = [[Lhs1, Rhs1]|New_out].

type_2_red([[Lhs1, Rhs1]|Tail],Lambdas,
		Out_list,Failure,Bad_rhs):-
	nos_non_zero_elems(Lhs1,0),      
	dot_prod(Rhs1, Lambdas, X),
	0 =< X,
	type_2_red(Tail, Lambdas, Out_list,Failure,Bad_rhs).

type_2_red([[Lhs1, Rhs1]|Tail],Lambdas,
		Out_list,true,Rhs1):-
	nos_non_zero_elems(Lhs1,0),      
	dot_prod(Rhs1, Lambdas, X),
	0 > X.


/******************************************************************

		FOURIER-MOTZKIN ELIMINATION
			PROCESS

*****************************************************************/

/****************************************************************

fm_r(+List_ineq, +Solution_vector, +Nos_elim, +Lambdas, 
	-Inconsistency, -Reduced_sys, -Var_left, -Bad_rhs)

*/

fm_r(List_ineq, [Head|[]], _, Lambdas, 
	false, List_ineq, Head, _).

fm_r(List_ineq, Solution_vector, Nos_elim, Lambdas, 
	Inconsistency, Reduced_sys, Var_left, Bad_rhs):-
        eliminate(List_ineq, Solution_vector, Curr_elim,
                Elim_list, New_vector),
        New_nos_elim is Nos_elim + 1,
        type_1_red(New_nos_elim, Elim_list, Elim1_list),
        type_2_red(Elim1_list, Lambdas, Elim2_list, Inc, Bad_rhs),
        if( Inc = false,
		(write_string("which reduces to:"), 
		nl, 
		write_sys_right(Elim2_list, New_vector, Lambdas),
		nl, 
		nl,
		fm_r(Elim2_list, New_vector, New_nos_elim, Lambdas,
			Inconsistency, Reduced_sys, Var_left, Bad_rhs)),
		(Inconsistency = Inc)).

/****************************************************************

fm(+List_ineq, +Solution_vector, +Lambdas, 
	-Inconsistency, -Reduced_sys, -Var_left, -Bad_rhs):-


*/

fm(List_ineq, Solution_vector, Lambdas, 
	Inconsistency, Reduced_sys, Var_left, Bad_rhs):-
	nl, 
	nl, 
	write_string("DECIDING WHETHER A SYSTEM IS CONSISTENT"), 
	nl,
	write_string("======================================="), 
	nl,
	nl, 
	nl, 
	write_string("Initially, the system looks like:"),
	nl, 
	write_sys_right(List_ineq, Solution_vector, Lambdas),
	nl, 
	nl, 
	type_2_red(List_ineq, Lambdas, New_list_ineq, Inc, Bad_rhs),
	if(Inc = false,
		fm_r(New_list_ineq, Solution_vector, 0, Lambdas, Inconsistency, 
		Reduced_sys, Var_left, Bad_rhs),
		Inc = Inconsistency).

/* ********************************************************
	CONCLUSION OF CODE FOR consistent
**********************************************************/




/*********************************************************
**********************************************************

PHASE 2 - REDUCTION COMPONENT
=============================

In this phase of the tactic the reduction is carried
out. Essentially this amounts to repeating the
following process until an inconsistency is 
deduced:

	i) Eliminate hypotheses that are redundant
	by virtue of their lhs being equivalent to
	zero. (==> test rhs for inconsistency here).

	ii) Eliminate a variable from the lhs of the
	system of inequalities being considered. Choose
	this variable "cleverly".

	iii) Eliminate hypotheses that are redundant
	because they are a consequence of other 
	hypotheses in the system.

	iv) If there is only one variable left, perform
	a consistency analysis on this.

red_fm(+Hyp_list, +Bad_rhs).
*/

red_fm(Hyp_list, Bad_rhs):-
	nl,
	write_string("COMBINATION PHASE"),
	nl,
	write_string("================="),
	nl,
	call_tac(combine(Hyp_list, Bad_rhs)),
	nl,
	write_string("SIMPLIFICATION PHASE"),
	nl,
	write_string("===================="),
	nl,
	call_tac(fm_simp),
	nl. 



/* ************************************************************

	THE COMBINATION TACTIC
	======================

*/

?- tactic(combine(Hyp_list, Bad_rhs),
	[],
        'Combining hypotheses to get contradiction',
        []).


combine([H1,H2],[1,C2]):-
        is_hyp(H1,Lhs1 =< Rhs1),
        is_hyp(H2,Lhs2 =< Rhs2),
        fwd(arith:linear1_comb, (Lhs1 + C2*Lhs2 =< Rhs1 + C2*Rhs2)).

combine([H1,H2],[C1,C2]):-
        is_hyp(H1,Lhs1 =< Rhs1),
        is_hyp(H2,Lhs2 =< Rhs2),
        fwd(arith:linear_comb, (C1*Lhs1 + C2*Lhs2 =< C1*Rhs1 + C2*Rhs2)).

combine([H1|Htail],[C1|Ctail]):-
	C1 = 0,
	combine(Htail,Ctail).

combine([H1|Htail],[C1|Ctail]):-
        Htail = [H2|H2tail],
        Ctail = [C2|C2tail],
        C2 = 0,
	New_hyp_list = [H1|H2tail],
	New_c_list = [C1|C2tail],
        combine(New_hyp_list,New_c_list).


combine([H1|Htail],[C1|Ctail]):-
        Htail = [H2|H2tail],
        Ctail = [C2|C2tail],
        combine([H1,H2],[C1,C2]),
        number_of_hyps(Hn),
        New_hyp_list = [Hn|H2tail],
        New_c_list = [1|C2tail],
        combine(New_hyp_list,New_c_list).

combine([H1],_).


/* ******************************************************

	THE SIMPLIFICATION PHASE
	========================


*/

?- tactic(fm_simp,
	[],
	'This tactic simplifies the combined inequalities',
	[]).


fm_simp:-
	number_of_hyps(N),
	focus((h N),
		(is_focus(Term),
		 call_tac(int_simp(Term)),
		 is_focus(Term1),
		 call_tac(int_simp(Term1)),
		 focus([1],
			(is_focus(T1),
			 call_tac(re_assoc_terms(T1)),
			 is_focus(T2),
			 call_tac(sort_and_simp(T2)),
			 is_focus(T3),
			 call_tac(add_like_terms(T3)),
			 is_focus(T4),
			 int_simp(T4))),
		 _)).


