%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS std_heur.ql,v 1.13 1993/12/10 02:19:10 marku Exp


%% Some standard heuristics for the 'r' theory.
%% (some are actually quite general, and could be global).

:- heuristic(split_hyp, 1, 
	'Splits hypotheses of the form (_ and _) into separate hypotheses',
	is_hyp(_, _ and _),
	[],
	split_all_hyps,
	'Splitting conjunct hypothesis').


:- heuristic(use_hyp, 1,
	'Looks for hypotheses that are identical to the focus or false',
	heur_use_hyp(Num,Hyp,Code),
	['Using hypothesis ',Num,': ',Hyp],
	Code,
	[]).

heur_use_hyp(Num, false, set_focus(G)) :-
    	is_hyp(Num, false),
	is_focus(F),
    	is_goal(G),
	\+ (F == G).

heur_use_hyp(Num, F, trans_true(h F)) :-
	is_focus(F),
	F \== true,
	is_hyp(Num,F).

heur_use_hyp(Num, not F, trans(r:notimpbool, [Num])) :-
	is_focus(F),
	F \== false,
	is_goal(false),
	is_hyp(Num, not F).



:- heuristic(conjuncts, 2,
	'Splits the focus into conjuncts',
	(is_focus(A and B),is_goal(true)),
	'Splitting the focus into conjuncts...',
	conjuncts(_),
	[]).
	
:- heuristic(when_focus(A, B), 2,
        ['when_focus(F,Tac) -- calls Tac when focus matches F'],
        (\+ \+ is_focus(A), functor(B, BF, _)),
        ['Trying ', BF, ' on a focus matching ', A],
	B,
        []).

:- heuristic(when(A, B), 2,
        ['when(Tst,Tac) -- calls Tac when Tst succeeds.',nl,
	 'Note that variables in Tst are not instantiated',nl,
         'at the time when Tac is called'],
        (\+ \+ A),
        ['Trying ', B],
        B,
        []).


:- heuristic(discharge_hyp, 2, 
	'Attempts to discharge conjectures',
	(\+ is_justified_focus),
	['Trying to discharge a conjecture'],
	discharge_hyp(N),
	['Discharged conjecture number ', N]).

:- heuristic(localness, 3, 
	'Attempts to use the localness rule group',
	(is_focus(is_local x _) ; is_focus(l_local x _)),
	['Trying the ''localness'' rule group'],
        rewrite_group(localness),
	[]).



:- heuristic(simplify_if, 1,
	'Uses basic laws about if(_,_,_) to simplify the focus',
	(is_focus(if(A,B,C)),ifprf(A,B,C,Msg,Code)),
	Msg,
	wrap_tac(Code),
	[]).


ifprf(true,  _, _,    'Simplifying if(true,...)', trans(r:iftrue)).
ifprf(false, _, _,    'Simplifying if(false,...)', trans(r:iffalse)).
ifprf(_, A, A,        'Simplifying if(_,X,X) to X', trans(r:ifparallel)).
ifprf(_, true, false, 'Simplifying if(X,true,false) to X',trans(r:if_true_false)).
ifprf(A, _, _,        'Using hypothesis to prove if condition', Code) :-
	 is_hyp(_, A),
	 Code = open_pos([1], [1], [], trans_true(h A)).



:- heuristic(if_parts, 3,
	'Tries focusing on the arguments of if(_,_,_)',
	(is_focus(if(A,B,C)),ifprf2(A,B,C,PreMsg,Code)),
	PreMsg,
	wrap_tac(Code),
	[]).

ifprf2(A,_,_, 'Focusing on `condition'' part of if(...)', Code) :-
	nonvar(A),
	reportWindow,
	Code = (focus([1], simp_if, _), \+ is_focus(if(A,_,_))).

ifprf2(_,A,_, 'Focusing on `then'' part of if(...)', Code) :-
        A \= true,
        A \= false,
	Code = (focus([2], simp_if, _), \+ is_focus(if(_,A,_))).

ifprf2(_,_,A, 'Focusing on `else'' part of if(...)', Code) :-
        A \= true,
        A \= false,
	Code = (focus([3], simp_if, _), \+ is_focus(if(_,_,A))).



:- heuristic(expand_def(Ops), 3,
	['Unfolds functions in the given list whenever they ',nl,
	 'appear at the outermost level of the focus'],
	(is_focus(F),functor(F,Op,_),elem(Op,Ops)),
	['Expanding the definition of `',Op,'''.'],
	trans(def(Op)),
	[]).


:- heuristic(always(L), 4,
	'Calls the given tactic unconditionally',
	true,
	['Calling tactic: ',L],
	L,
	[]).

:- heuristic(quit, 1, 
       'Quit from current window when focus == goal',
       (is_focus(F),
	is_goal(G),
	F==G,
	is_justified_focus,
	complete_retry_delays),
       ['Quit, because focus == goal  (',F,')'],
       quit,
       []).

/*
**  Some heuristic environments
*/

:- envir(empty,   2, []).
:- envir(std,     2, [[quit,use_hyp],
	              [conjuncts],
		      [localness, split_hyp],
		      [discharge_hyp, always(use(true,goal))]
		     ]).
:- envir(quit,    1, [[always(quit)]]).
:- envir(suspend, 1, [[always(do_suspend)]]).
:- envir(if_primitive, 2,
     [
         [use_hyp, split_hyp],
	 [simplify_if],
	 [expand_def([and, or, => , <=>, not]), if_parts],
	 [always(quit)]
     ]).
